/***************************************************************************
                          mccluskey.cpp  -  description
                             -------------------
    begin                : Sat Oct 6 2001
    copyright            : (C) 2001 by Dannel Albert
    email                : dalbert@capitol-college.edu
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include <cmath>
#include <strstream>
#include <iomanip>
#include <cassert>
#include "mccluskey.h"

using namespace std;

McCluskey::McCluskey(Quine* qq)
{
  q = qq;
  primeImplicants = q->GetPrimeImplicants(numPI);
  cdl = q->GetCanonicalDecimalList();
  piTable = new bool*[numPI];
  assert(piTable);
  for (unsigned int i=0; i<numPI; i++)
  {
    piTable[i] = new bool[cdl->GetNumCares()];
    assert(piTable[i]);
    for (unsigned int j=0; j<cdl->GetNumCares(); j++)
    {
      piTable[i][j] = false;
    }
  }
  booleanEq = "";
  numImplicants = 0;
  implicants = NULL;
}

McCluskey::~McCluskey()
{
  if (piTable)
  {
    for (unsigned int i=0; i<numPI; i++)
    {
      delete[] piTable[i];
    }
    delete[] piTable;
    piTable = NULL;
  }
}

void McCluskey::Solve()
{
  unum* cares;
  unsigned int nCares;

  cares = cdl->GetCares(nCares);
  for (unsigned int j=0; j<nCares; j++)
  {
    for (unsigned int i=0; i<numPI; i++)
    {
      for (unsigned int k=0; k<primeImplicants[i]->Size(); k++)
      {
        if (primeImplicants[i]->ValueAt(k) == cares[j])
          piTable[i][j] = true;
      }
    }
  }

  FewestImplicants();
  numImplicants = imp.Size();
  implicants = new unum[numImplicants];
  assert(implicants);
  for (unsigned int i=0; i<numImplicants; i++)
    implicants[i] = imp.Pop();

  CreateBooleanEquation();
}

void McCluskey::FewestImplicants()
{
  unsigned int caresCovered=0;
  unsigned int nCares;
  unum* cares;
  UNumStack impCover;
  bool* covered;
  unsigned int numWantCovered=1;

  cares = cdl->GetCares(nCares);
  covered = new bool[nCares];
  assert(covered);
  for (unsigned int i=0; i<nCares; i++)
    covered[i] = false;

  while(caresCovered != nCares)
  {
    for (unsigned int i=0; i<nCares; i++)
    {
      // clear out the stack
      while (impCover.IsEmpty() == false)
        impCover.Pop();

      // count how many implicants cover this group
      for (unsigned int j=0; j<numPI; j++)
      {
        if (piTable[j][i] == true && covered[i] == false)
          impCover.Push(j);
      }

      if (impCover.Size() == numWantCovered)
      {
        if (impCover.Size() == 1)
        {
          imp.Push(impCover.Pop());

          // find what cares it covers
          for (unsigned int l=0; l<nCares; l++)
          {
            if (piTable[imp.Peek()][l] == true && covered[l] == false)
            {
              covered[l] = true;
              caresCovered++;
            }
          }
        }
        else
        {
          unsigned int bestImp = impCover.Peek();
          unsigned int thisImp;
          unsigned int nBestCovered = 0;

          // find out which of these implicants
          // covers the most other uncovered groups
          while (impCover.IsEmpty() == false)
          {
            thisImp = impCover.Pop();
            unsigned int cCares=0;
            for (unsigned int ii=0; ii<nCares; ii++)
            {
              if (piTable[thisImp][ii] == true && covered[ii] == false)
              {
                cCares++;
              }
            }

            if (cCares > nBestCovered)
            {
              nBestCovered = cCares;
              bestImp = thisImp;
            }
          }

          imp.Push(bestImp);

          // find what cares it covers
          for (unsigned int l=0; l<nCares; l++)
          {
            if (piTable[imp.Peek()][l] == true && covered[l] == false)
            {
              covered[l] = true;
              caresCovered++;
            }
          }
        }
      }
    }

    numWantCovered++;
  }

  delete[] covered;
  covered = NULL;
}

void McCluskey::CreateBooleanEquation()
{
#define WHITESPACE " "
#define NOT "'"
#define AND ""
#define OR "+"
  unum value, mask;
  unsigned int* diffBits;
  unsigned int numDiffBits;
  unsigned int numBits;
  std::string* inputs;
  std::strstream beq;

  inputs = cdl->GetInputs(numBits);

  for (int i=(int)numImplicants-1; i>=0; i--)
  {
    value = primeImplicants[implicants[i]]->ValueAt(0);
    diffBits = primeImplicants[implicants[i]]->GetDiffBits(numDiffBits);

    mask = (unum)pow(2,numBits-1);
    for (int j=(int)numBits-1; j>=0; j--)
    {
      bool dBit=false;

      for (unsigned int k=0; k<numDiffBits; k++)
      {
        if (diffBits[k] == mask)
          dBit = true;
      }

      if (dBit == false)
      {
        if ((value & mask) == mask)
          booleanEq += inputs[numBits-j-1];
        else
          booleanEq += inputs[numBits-j-1] + NOT;

        if (j != 0)
        {
          booleanEq += AND;
        }
      }

      mask = mask >> 1;
    }

    if (i != 0)
    {
      booleanEq += WHITESPACE;
      booleanEq += OR;
      booleanEq += WHITESPACE;
    }
  }

#undef WHITESPACE
#undef AND
#undef OR
#undef NOT
}

ostream& operator<<(std::ostream& outs, const McCluskey& mc)
{
#define CELL_WIDTH 4
#define SIDE_WIDTH 6
#define WHITESPACE " "
#define MARKER "X"
#define IMPLICANT "*"
  unsigned int i;
  cout << "[PI]" << endl;
  for (unsigned int i=0; i<mc.numPI; i++)
  {
    outs << *mc.primeImplicants[i] << endl;
  }

  if (mc.piTable)
  {
    unum* cares;
    unsigned int nCares;

    string str;

    cares = mc.cdl->GetCares(nCares);
    outs << endl << WHITESPACE << setw(SIDE_WIDTH) << WHITESPACE <<
          setw(CELL_WIDTH) << WHITESPACE;
    for (i=0; i<nCares; i++)
    {
      outs << setw(CELL_WIDTH) << cares[i] << WHITESPACE;
    }
    outs << endl;

    for (i=0; i<mc.numPI; i++)
    {
      if (mc.implicants)
      {
        unsigned int index;
        for (index=0; index<mc.numImplicants &&
            mc.implicants[index] != i; index++)
        {}
        if (mc.implicants[index] == i)
          outs << IMPLICANT;
        else
          outs << WHITESPACE;
      }
      else
        outs << WHITESPACE;

      outs << setw(SIDE_WIDTH) << i+1 <<
            setw(CELL_WIDTH) << WHITESPACE;
      for (unsigned int k=0; k<mc.cdl->GetNumCares(); k++)
      {
        if (mc.piTable[i][k] == true)
          outs << setw(CELL_WIDTH) << MARKER;
        else
          outs << setw(CELL_WIDTH) << WHITESPACE;
        outs << WHITESPACE;
      }
      outs << endl;
    }
  }

  string* inputs;
  unsigned int numInputs;
  inputs = mc.cdl->GetInputs(numInputs);
  outs << endl << "f(";
  for(i=0; i<numInputs; i++)
  {
    outs << inputs[i];
    if (i != numInputs-1)
      outs << ",";
  }
  outs << ") = ";
  outs << mc.booleanEq << endl;

  return outs;
#undef CELL_WIDTH
#undef SIDE_WIDTH
#undef WHITESPACE
#undef MARKER
#undef IMPLICANT
}