/***************************************************************************
                          quine.cpp  -  description
                             -------------------
    begin                : Sat Oct 6 2001
    copyright            : (C) 2001 by Dannel Albert
    email                : dalbert@capitol-college.edu
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include <cassert>
#include "quine.h"

using namespace std;

unum Sum(const QuineGroup* qg);
QuineGroup* GroupIsThere(const QuineGroup* qg1, const QuineGroup* qg2,
                  const QuineTerm* qt);
bool SameDiffBits(const QuineGroup* qg1, const QuineGroup* qg2);

Quine::Quine()
{
  head = tail = NULL;

  cdl = new CanonicalDecimalList;
  assert(cdl);

  primeImplicants = NULL;
  numSteps = 0;
  numInputs = 0;
  numPI = 0;
  solved = false;
}

Quine::~Quine()
{
  QuineList* nd = head;
  QuineList* trash;
  while(nd != NULL)
  {
    trash = nd;
    nd = nd->GetNext();
    delete trash;
  }

  if (primeImplicants)
  {
    delete[] primeImplicants;
    primeImplicants = NULL;
  }

  if (cdl)
  {
    delete cdl;
    cdl = NULL;
  }

  numSteps = 0;
  numInputs = 0;
  numPI = 0;
  solved = false;
  head = NULL;
}

bool Quine::Solve()
{
  if (solved)
    return false;
  if (!Setup())
    return false;

  bool done=false, combined=false;
  QuineList* newList;
  QuineList* curList = head;
  QuineTerm* newTerm;
  QuineTerm* curTerm;
  QuineGroup* group1;
  QuineGroup* group2;
  QuineGroup* combo;
  unum sumG, sumNG;
  numPI = 0;

  do
  {
    unsigned int i;
    combined = false;

    // insert new list
    newList = new QuineList;
    assert(newList);
    Insert(newList);

    // insert new terms
    for (i=0; i<curList->Size()-1; i++)
    {
      newTerm = new QuineTerm;
      assert(newTerm);
      newList->Insert(newTerm);
    }

    for (unsigned int k=0; k<curList->Size()-1; k++)
    {
      newTerm = newList->TermAt(k);
      for (i=0; i<curList->TermAt(k)->Size(); i++)
      {
        group1 = curList->TermAt(k)->GroupAt(i);
        group1->SetPrimeImplicant(numPI+1);
        sumG = Sum(group1);
        for (unsigned int j=0; j<curList->TermAt(k+1)->Size(); j++)
        {
          group2 = curList->TermAt(k+1)->GroupAt(j);
          sumNG = Sum(group2);
          if (SameDiffBits(group1, group2) &&
              (sumNG > sumG) && PowerOf2(sumNG - sumG, numInputs))
          {
            combined = true;
            group1->SetHasCombined(true);
            group1->SetPrimeImplicant(0);
            group2->SetHasCombined(true);
            combo = GroupIsThere(group1, group2, newTerm);
            if (combo == NULL)
            {
              // create the new group
              combo = new QuineGroup(numInputs);
              assert(combo);

              for (unsigned int ii=0; ii<group1->Size(); ii++)
              {
                combo->Insert(group1->ValueAt(ii));
                combo->Insert(group2->ValueAt(ii));
              }
              combo->CalculateDiffBits();

              // assign it to the term
              newTerm->Insert(combo);
            }
            else
            {
              // increment repeatCount
              combo->SetRepeatCount(combo->GetRepeatCount()+1);
            }
          }
        }

        if (group1->HasCombined() == false)
        {
          numPI++;
        }
      }
    }

      // check the last term for prime implicants
      unsigned int iLast=curList->Size()-1;
      for (i=0; i<curList->TermAt(iLast)->Size(); i++)
      {
        group1 = curList->TermAt(iLast)->GroupAt(i);
        if (group1->HasCombined() == false)
        {
          group1->SetPrimeImplicant(++numPI);
        }
      }

    curList = newList;
    curTerm = newList->TermAt(0);
    if (combined == false)
    {
      done = true;
    }
  }
  while (!done);

  CreatePIList();

  solved = true;
  return true;
}

bool Quine::Setup()
{
  QuineGroup** qgCares;
  QuineGroup** qgDontCares;
  QuineGroup* qg;
  QuineTerm* qt;
  unum* cares;
  unum* dontCares;
  unum value, test;
  unsigned int numCares, numDontCares, num1s;
  unsigned int i;

  Insert(new QuineList);

  numInputs = cdl->GetNumInputs();
  if (numInputs == 0)
    return false;
  cares = cdl->GetCares(numCares);
  dontCares = cdl->GetDontCares(numDontCares);

  if (!cares)
    return false;

  qgCares = new QuineGroup*[numCares];
  assert(qgCares);
  qgDontCares = new QuineGroup*[numDontCares];
  assert(qgDontCares);

  // put values into groups (each value is considered a group)
  for (i=0; i<numCares; i++)
  {
    qg = new QuineGroup(numInputs);
    assert(qg);
    qgCares[i] = qg;
    qgCares[i]->Insert(cares[i]);
  }
  for (i=0; i<numDontCares; i++)
  {
    qg = new QuineGroup(numInputs);
    assert(qg);
    qgDontCares[i] = qg;
    qgDontCares[i]->Insert(dontCares[i]);
  }

  // create the QuineTerms
  for (i=0; i<numInputs+1; i++) // numInputs+1 b/c 0 is a term
  {
    qt = new QuineTerm;
    assert(qt);
    head->Insert(qt);
  }

  // put groups into terms
  // cares
  for (i=0; i<numCares; i++)
  {
    value = cares[i];
    test = value;
    num1s = 0;
    for (unsigned int j=0; j<numInputs; j++)
    {
      if ((test & 1) == 1)
        num1s++;
      test = test >> 1;
    }

    // the number of 1s in the decimal value determine what term it belongs to
    qt = head->TermAt(num1s);
    qt->Insert(qgCares[i]);
  }

  // dont cares
  for (i=0; i<numDontCares; i++)
  {
    value = dontCares[i];
    test = value;
    num1s = 0;
    for (unsigned int j=0; j<numInputs; j++)
    {
      if ((test & 1) == 1)
        num1s++;
      test = test >> 1;
    }

    // the number of 1s in the decimal value determine what term it belongs to
    qt = head->TermAt(num1s);
    qt->Insert(qgDontCares[i]);
  }

  delete[] qgCares;
  qgCares = NULL;
  delete[] qgDontCares;
  qgDontCares = NULL;

  return true;
}

void Quine::CreatePIList()
{
  QuineList* ql = head;
  QuineTerm* qt;
  QuineGroup* qg;
  unsigned int i=0;

  primeImplicants = new QuineGroup*[numPI];
  assert(primeImplicants);

  while (ql != NULL)
  {
    qt = ql->TermAt(0);
    while (qt != NULL)
    {
      qg = qt->GroupAt(0);
      while (qg != NULL)
      {
        if (!qg->HasCombined() && qg->PrimeImplicant() != 0)
        {
          primeImplicants[i] = qg;
          i++;
        }
        qg = qg->GetNext();
      }
      qt = qt->GetNext();
    }
    ql = ql->GetNext();
  }
}

void Quine::Insert(QuineList* ql)
{
  if (!numSteps)
  {
    head = ql;
    assert(head);
    tail = head;
  }
  else // insert at the end of the list
  {
    ql->SetPrev(tail);
    tail->SetNext(ql);
    tail = ql;
  }

  numSteps++;
}

ostream& operator<<(ostream& outs, const Quine& q)
{
  QuineList* ql = q.head;
  for (unsigned int i=0; i<q.numSteps; i++)
  {
    outs << *ql;
    if (i != q.numSteps-1)
      outs << "." << endl;
    ql = ql->GetNext();
  }
  return outs;
}

istream& operator>>(istream& ins, Quine& q)
{
  ins >> *q.cdl;
  return ins;
}

unum Sum(const QuineGroup* qg)
{
  unum s=0;
  for (unsigned int i=0; i<qg->Size(); i++)
  {
    s += qg->ValueAt(i);
  }
  return s;
}

QuineGroup* GroupIsThere(const QuineGroup* qg1, const QuineGroup* qg2,
                  const QuineTerm* qt)
{
  QuineGroup* qg;
  QuineGroup temp(qg1->GetNumBits());

  // combine the two groups into one group, temporarily
  // it is assumed that the two groups have the same size
  for (unsigned int k=0; k<qg1->Size(); k++)
  {
    temp.Insert(qg1->ValueAt(k));
    temp.Insert(qg2->ValueAt(k));
  }

  for (unsigned int i=0; i<qt->Size(); i++)
  {
    // go through each group in the term
    qg = qt->GroupAt(i);
    for (unsigned int j=0; j<qg->Size(); j++)
    {
      // see if their values are equal to those in our temp
      unsigned int k=0;
      while(k < qg->Size() && qg->ValueAt(k) == temp.ValueAt(k))
      { k++; }

      if (k == qg->Size())
      {
        // found a match
        return qg;
      }
    }
  }

  return NULL;
}

bool SameDiffBits(const QuineGroup* qg1, const QuineGroup* qg2)
{
  unum* db1, *db2;
  unsigned int nDb1, nDb2;

  db1 = qg1->GetDiffBits(nDb1);
  db2 = qg2->GetDiffBits(nDb2);

  if (nDb1 == 0)
    return true;

  unsigned int i=0;
  while (i < nDb1 && db1[i] == db2[i])
  { i++; }

  if (i == nDb1)
    return true;

  return false;
}